import pathlib
import typing

from aemeasure import Measurement
from aemeasure.database import Database


class MeasurementSeries:
    """
    A series of measurements.
    """

    def __init__(self, db: typing.Union[Database, str, pathlib.Path],
                 stdout: typing.Optional[str] = "stdout",
                 stderr: typing.Optional[str] = "stderr",
                 metadata: bool = True,
                 cache: bool = True):
        """
        By default, the series will save
        :param path: Path to the database or database itself.
        :param stdout: Capture stdout under this label for every measurement. Set to
                        `None` if you don't want this feature.
        :param stderr: Capture stderr under this label for every measurement. Set to
                        `None` if your don't want this feature.
        :param metadata: Automatically save metadata.
        :param cache: Wait until the end of the series to flush the database to disk.
        """
        self.db = db if isinstance(db, Database) else Database(db)
        self.cache = cache
        self._stderr = stderr
        self._stdout = stdout
        self._save_metadata = metadata

    def measurement(self, cache: typing.Optional[bool] = None) -> Measurement:
        """
        Create a new measurement. A measurement should be for example
        an algorithm solving a single instance. Note that this is not
        meant for micro-benchmarks: If you try to measure running times
        significantly below 1s, you should use another tool.
        capture_stdout: Saves the output of stdout into this field. Note that
        saving the output can use up a significant amount of disk storage.
        capture_stderr: Saves the output of stderr into this field.
        :param cache: If set to true, the entry will only be saved to disk when
        the datbase is flushed or closed. If set to false, every measurement
        will directly be written to disk. If not set, the top level decision
        is used.
        """
        cache = self.cache if cache is None else cache
        return Measurement(self.db,
                           capture_stdout=self._stdout,
                           capture_stderr=self._stderr,
                           save_metadata=self._save_metadata,
                           cache=cache)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.db.flush()



import subprocess


def get_git_revision():
    try:
        label = (
            subprocess.check_output(["git", "rev-parse", "HEAD"])
            .strip()
            .decode("ascii")
        )
    except subprocess.CalledProcessError:
        print("Warning: Could not read git-revision!")
        label = None
    return label



import datetime
import json
import logging
import os.path
import pathlib
import random
import socket
import typing
import zipfile
from zipfile import ZipFile

_log = logging.getLogger("AeMeasure")


class Database:
    """
    A simple database to dump data (dictionaries) into. Should be reasonably threadsafe
    even for slurm pools with NFS.
    """

    def __init__(self, path: typing.Union[str, pathlib.Path]):
        self.path = path
        if not os.path.exists(path):
            # Could fail in very few unlucky cases on an NFS (parallel creations)
            os.makedirs(path, exist_ok=True)
            _log.info(f"Created new database '{path}'.")
        if os.path.isfile(path):
            raise RuntimeError(
                f"Cannot create database {path} because there exists an equally named file."
            )
        self._subfile_path = self._get_unique_name()
        self._cache = []

    def _get_unique_name(self, __tries=3):
        """
        Generate a unique file name to prevent collisions of parallel processes.
        """
        if __tries <= 0:
            raise RuntimeError("Could not generate a unique file name. This is odd.")
        hostname = socket.gethostname()
        timestamp = datetime.datetime.now().strftime("%Y%m%d%H%M")
        rand = random.randint(0, 10000)
        name = f"{timestamp}-{hostname}-{rand}.data"
        if os.path.exists(name):
            return self._get_unique_name(__tries=__tries - 1)
        return name

    def compress(self, compression=zipfile.ZIP_LZMA, compresslevel=None):
        """
        Warning: This may not be threadsafe! If you want to extract all data to
        a single file, just use 'read' and dump the output into a single json.
        """
        compr_path = os.path.join(self.path, "_compressed.zip")
        with ZipFile(compr_path, "a", compression=compression,
                     compresslevel=compresslevel) as z:
            for file_name in os.listdir(self.path):
                path = os.path.join(self.path, file_name)
                if not os.path.isfile(path) or not path.endswith(".data"):
                    continue
                if os.path.getsize(path) <= 0:
                    _log.warning(f"Skipping '{path}' due to zero size.")
                    continue
                _log.info(f"Compressing '{file_name}' of size {os.path.getsize(path)}.")
                z.write(path, file_name)
                os.remove(path)
        _log.info(f"Compressed database has size {os.path.getsize(compr_path)}.")

    def dump(self, entries: typing.List[typing.Dict], flush=True):
        if isinstance(entries, dict):
            raise ValueError("Use 'add' to dump a single dictionary.")
        self._cache += entries
        if flush:
            self.flush()

    def add(self, entry: typing.Dict, flush=True):
        self.dump([entry], flush)

    def flush(self):
        if not self._cache:
            return
        with open(os.path.join(self.path, self._subfile_path), "a") as f:
            for data in self._cache:
                f.write(json.dumps(data) + "\n")
            _log.info(f"Wrote {len(self._cache)} entries to disk.")
            self._cache.clear()

    def load(self) -> typing.List[typing.Dict]:
        data = list(self._cache)
        # load compressed data
        compr_path = os.path.join(self.path, "_compressed.zip")
        if os.path.exists(compr_path):
            with ZipFile(compr_path, "r") as z:
                for filename in z.filelist:
                    with z.open(filename, "r") as f:
                        for line in f.readlines():
                            data.append(json.loads(line))
        # load uncompressed data
        for fp in os.listdir(self.path):
            path = os.path.join(self.path, fp)
            if not os.path.isfile(path) or not path.endswith(".data"):
                continue
            with open(path, "r") as f:
                for entry in f.readlines():
                    data.append(json.loads(entry))
        return data

    def clear(self):
        """
        Clear database (cache and disk). Note that remaining data in the
        cache of other nodes may still be written.
        """
        # cache
        self._cache.clear()
        # compressed
        compr_path = os.path.join(self.path, "_compressed.zip")
        if os.path.exists(compr_path):
            os.remove(compr_path)
        # remaining .data files
        for fp in os.listdir(self.path):
            path = os.path.join(self.path, fp)
            if not os.path.isfile(path) or not str(path).endswith(".data"):
                continue
            os.remove(path)

    def __del__(self):
        self.flush()



import pathlib
import typing

import pandas as pd

from aemeasure import Database


def data_to_pandas(
    data: typing.List[typing.Dict], defaults: typing.Optional[typing.Dict] = None
):
    if not defaults:
        defaults = {}
    # find columns
    for entry in data:
        for key in entry:
            if key not in defaults:
                defaults[key] = None
    pd_data = {key: [] for key in defaults}

    for entry in data:
        entry_ = dict(defaults)
        entry_.update(entry)
        for key, value in entry_.items():
            pd_data[key].append(value)

    return pd.DataFrame(pd_data)


def read_as_pandas_table(
    path: typing.Union[str, pathlib.Path], defaults: typing.Optional[typing.Dict] = None
):
    db = Database(path)
    data = db.load()
    return data_to_pandas(data, defaults)



import datetime
import os
import socket
import sys
import typing

from .database import Database
from .utils.capture import OutputCopy
from .utils.git import get_git_revision


class Measurement(dict):
    _measurement_stack = []
    _git_revision = get_git_revision()

    @staticmethod
    def last() -> dict:
        return Measurement._measurement_stack[-1]

    def time(self, timer=None) -> datetime.timedelta:
        if timer is None:
            return datetime.datetime.now() - self._time
        else:
            return datetime.datetime.now() - self._timer[timer]

    def save_timestamp(self, key="timestamp"):
        self[key] = datetime.datetime.now().isoformat()

    def save_seconds(self, key="runtime", timer=None):
        v = self.time(timer).total_seconds()
        self[key] = v
        return v

    def save_hostname(self, key="hostname"):
        v = socket.gethostname()
        self[key] = v
        return v

    def save_argv(self, key="argv"):
        if sys.argv:
            v = " ".join(sys.argv)
            self[key] = v
            return v
        else:
            self[key] = None
            return None

    def save_git_revision(self, key="git_revision") -> str:
        self[key] = self._git_revision
        return self._git_revision

    def save_metadata(self):
        self.save_seconds()
        self.save_timestamp()
        self.save_hostname()
        self.save_argv()
        self.save_git_revision()
        self.save_cwd()

    def start_timer(self, name: str):
        self._timer[name] = datetime.datetime.now()

    def __init__(
            self,
            db: typing.Union[Database, str],
            capture_stdout: typing.Optional[str] = None,
            capture_stderr: typing.Optional[str] = None,
            save_metadata = False,
            cache=False,
    ):
        super().__init__()
        self._time = datetime.datetime.now()
        if isinstance(db, Database):
            self._db = db
        else:
            self._db = Database(db)
        self._timer = dict()
        self._capture_stdout = capture_stdout
        self._capture_stderr = capture_stderr
        self._cache = cache
        self._save_metadata = save_metadata

    def __enter__(self):
        self._measurement_stack.append(self)
        self.start_capture()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop_capture()
        if self._save_metadata:
            self.save_metadata()
        if not exc_type:
            self.write()
        else:
            print("Do not save measurement due to exception.")
        e = self._measurement_stack.pop()
        assert e is self

    def save_cwd(self, label="cwd"):
        self[label] = os.getcwd()

    def start_capture(self):
        if self._capture_stdout:
            sys.stdout = OutputCopy(sys.stdout)
        if self._capture_stderr:
            sys.stderr = OutputCopy(sys.stderr)

    def stop_capture(self):
        if self._capture_stdout:
            self[self._capture_stdout] = sys.stdout.getvalue()
            sys.stdout = sys.stdout.wrapped_stream
        if self._capture_stderr:
            self[self._capture_stderr] = sys.stderr.getvalue()
            sys.stderr = sys.stderr.wrapped_stream

    def write(self):
        self._db.dump([self], self._cache)
